;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                             ;
; vm86.asm: implementation of code to switch out of virtual 8086 mode         ;
;             (Adapted from Loadlin Code by H. Lerman)                                                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;   This program is free software; you can redistribute it and/or modify
;   it under the terms of the GNU General Public License as published by
;   the Free Software Foundation; either version 2 of the License, or
;   (at your option) any later version.
;
;   This program is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License
;   along with this program; if not, write to the Free Software
;   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; This code is necessary in order for mloader to boot Minix starting from
; virtual 8986 mode. This happens, e.g., when the user invokes wini from
; the Dos command line while an extended memory manager, such as EMM386.EXE,
; is running. Many people don't realize it, but in this situation the DOS
; session is running "under" EMM386 in vm86 mode, much as it does in a 
; box under Windows. Most of the EMM386 code runs in protected mode. Thus,
; in effect, Emm386 is a primitive tasking operating system. Actually, a more
; apt description is "server", since Emm386 only runs at the request of
; user programs invoking service via INT 67H. 
;
; Most of the code in this file is taken, with some adaptations, from
; the Linux Loadlin code, written by Hans Lermen. Lermen refers to the
; API offered to real-mode programs by Emm386  the "VCPI server", for
; Virtual Command Program Interface. For more information, see the L.I.M.
; EMS API descriptions under INT 67H in Ralf Brown's interrupt list. (The
; "Bible" of Dos hackers everywhere.)
;
; The goal of this code is to implement a function _exit_vm86(). The effect
; of calling this routine from a program running in virtual 8086 mode is to
; transfer control to the following instruction, but with the CPU switched to
; real mode, all registers and stack being preserved.
;
; Note that after running exit_vm86 it will probably not be possible to
; run any DOS services, particularly if DOS had been loaded high with a
; DOS=UMB statement in config.sys.
;
; The strategy for this is the following: Use int 67h service 0DE0Ch to
; jump to protected mode. This service enables the user to designate a task
; which will be run from protected mode. The function of the designated task
; is merely to switch the processor to real mode. Of course, that word
; "merely" covers up a lot of things, like setting up the GDT for the jump
; to protected mode and loading the segment registers properly for the return
; to real mode.
;
; For an introduction to protected mode, and for a different method for
; switching modes from the one used here, consult "Protected Mode Basics",
; by Robert Collins, http://www.x86.org/articles/pmbasics/tspec_al_doc.html 
;
; Terry R. McConnell (tmc@barnyard.syr.edu)
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



.386
PUBLIC _check_VCPI_present, _exit_vm86
EXTRN _debug

align_  macro   start,val
        org     (((($-start)+(val-1))/val)*val)
        endm

DGROUP GROUP _DATA

_DATA SEGMENT WORD PUBLIC 'DATA' USE16
_GET_VERSION = 46h
_DEALLOCATE_PAGES = 45h
_ALLOCATE_RAW_PAGES = 5A01h
emm_int = 67h
return_value dw ?
handle dw ?
_DATA ENDS

CODE SEGMENT MEMPAGE PRIVATE 'Code' USE16
ASSUME cs:CODE, ds:CODE, ss:DGROUP
org 0
code_org0     label   byte

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                          ;
;                              DATA                                        ;
;                    (Address using CS register)                           ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

mystring0 db 'Finishing initializations',13,10,'$'
mystring1 db 'About to get vcpi interface',13,10,'$'
mystring2 db 'About to run switch_to ...',13,10,'$'
mystring3 db 'Back from protected mode: in vm86.asm',13,10,'$'
mystring4 db 'About to go to protected mode',13,10,'$'
mystring5 db 'Fell through switch to protected mode',13,10,'$'
mystring6 db 'Get VCPI interface failed ',13,10,'$'

; Macro used for printing a string from a location in the code for debugging
; purposes. Call as debugout memaddress, where memaddress is the address of
; of a string. Assumes ds = cs.

version m510
debugout macro outstring
	cmp debug_flag,0
	jz @F
	push ax
	push dx
	mov dx, offset outstring
	mov ah,9   ; Dos printstring
	int 21h
	pop dx
	pop ax
@@:
endm

C_SEG dw ?   ; the current value of cs ( = "HighSeg" in Lerman's code )
D_SEG dw ?       ; Space for stored values of segment registers
E_SEG dw ?
S_SEG dw ?
C_FRAME dd 0 ; C_SEG << 4: used to compute linear addresses ("High_Addr in
                                                         ;  Lerman's code.)
debug_flag dw 0  

; The structure of a descriptor: these are the entries in the GDT and LDT
; (= Global Descriptor Table, and Local Descriptor Table). (There are
; actually several kinds of descriptors. The following is the format of
; a segment descriptor.)
;
  descript    struc
    limit     dw    ?
    base0     dw    ?
    base16    db    ?
    typbyte   db    ?
    limit16   db    ?
    base24    db    ?
  descript    ends
;
; The following is the format of a call gate descriptor
; Not used in this code, but here in case we do later

  Gdescript struc
    gateoffs0    dw      ?
    gateselector dw      ?
    gatenotused  db      ?
    gatetyppbyte db      ?
    gateoffs16   dw      ?
  Gdescript ends


  ; definition of COMMON descriptor types (bit0..4 of descript.typbyte )
  ; (bit 4 of descript.typbyte =0)
  data_d        =     10000b    ; data segment descriptor
    writable    =     00010b    ; =1 if write acces allowed to data segment
    expand_down =     00100b    ; =1 limit counts down from base
  code_d        =     11000b    ; code segment
    readable    =     00010b    ; =1 if code also can be read (cannot be ovwritten)
    conforming  =     00100b    ; =1 code can be accesses and executed
                                ;    regardless of it's privilege level

  ; definition of SYSTEM decriptor types  (bit0..4 of descript.typbyte )
  ; (bit 4 of descript.typbyte =0)
  TSS286_avail_d   =  01h
  LDT_d            =  02h
  TSS286_busy_d    =  03h
  call_gate_d      =  04h
  task_gate_d      =  05h
  INT286_gate_d    =  06h
  TRAP286_gate_d   =  07h
  TSS386_avail_d   =  09h
  TSS386_busy_d    =  0bh
  call386_gate_d   =  0ch
  INT386_gate_d    =  0eh
  TRAP386_gate_d   =  0fh

  ; definition of privilege levels (bit5..6 of descript.typbyte )
  p0               =  0      ;
  p1               =  1*32   ; super visor levels
  p2               =  2*32   ;
  puser            =  3*32


  ;definition of granularity ( bits7..8 in descript.limit16 )
  gran_byte        =  0
  gran_page        =  10000000b  ; 4k granularity
  ; for data_selectors:
  data_USE16       =  0
  gran_big         =  01000000b  ; big segment
  data_USE32       =  gran_big   ; use 32-bit stack pointer ESP instead of SP
  ;                Intel says: relevant only together with expand_down  for data_d
  ;                But that is WRONG :
  ;                BIG segment must be set also if the descriptor is greater 64K
  ;                and is used to load SS !
  ;                (because SP cannot access behind 64K)
  ;
  ; for code_selectors:
  code_USE32       =  01000000b  ; default operand size 32 bit (for code segments)
  code_USE16       =  00000000b  ; default operand size 16 bit (for code segments)


  ; segment present bit (bit7 of descript.typbyte )
  is_present       =128
  not_present      =0

  descriptor macro name,typ,plevel,present,limit,gran,base
  name descript <limit and 0ffffh,base and 0ffffh,low (base shr 16),typ or plevel or present,(limit shr 16) or gran,high (base shr 16)>
            endm

  ; The above macro won't finish the job for us, because we cannot
  ; completely initialize the descriptors at assembly time. We need to know
  ; the value of the cs for this code segment. The following is used to
  ; fix this up at run time:

  fixup_descriptor macro name,base
	mov eax,base
	mov name.base0,ax
	shr eax,16
	mov name.base16,al
	mov name.base24,ah
	endm


  ;GDT Global Descriptor Table -------------------------v
  ; The base field is nonsense -- it is fixed in _exit_vm86 using the
  ; fixup_descriptor macro above
             align_  code_org0,16
  gdtnull    descript  <?>    ;0000   never accessable
  gdtvcpi_code descript <?>   ;0008
  gdtvcpi2   descript  <?>    ;0010
  gdtvcpi3   descript  <?>    ;0018
  descriptor gdt_core,(data_d+writable),p0,is_present,0fffffh,(gran_page+data_USE32),0
  descriptor gdt_code,(code_d+readable),p0,is_present,0ffffh,(gran_byte+code_USE16),0
  descriptor gdt_data,(data_d+writable),p0,is_present,0ffffh,(gran_byte+data_USE16),0
  descriptor gdt_ldt,LDT_d,p0,is_present,7,gran_byte,0
  descriptor gdt_tss,TSS286_avail_d,p0,is_present,0ffh,gran_byte,0
  gdtlast    descript  <?>    ; dummy for addressing

  g_vcpi_code equ      (gdtvcpi_code-gdtnull)
  g_core     equ       (gdt_core-gdtnull)
  g_code     equ       (gdt_code-gdtnull)
  g_data     equ       (gdt_data-gdtnull)
  g_ldt      equ       (gdt_ldt-gdtnull)
  g_tss      equ       (gdt_tss-gdtnull)
  ;GDT Global Descriptor Table -------------------------^


  ;LDT Local  Descriptor Table -------------------------v
  ldtnull    descript  <?>    ;0000   never accessable
  ldtlast    descript  <?>    ; dummy for addressing
  ;LDT Local  Descriptor Table -------------------------^

  ;                  align_  SYSTEMDATA_,1024
  ;IDT Interrupt Descriptor Table -------------------------v
  idtnull    descript  32 dup (<0>)
  idtlast    descript  <?>    ; dummy for addressing
  ;IDT Interrupt Descriptor Table -------------------------^

  our_tss    dd        128 dup (?)    ; our TSS Task State Segment


  ;params for switching TO protected mode -------------------------v
  ; Int 67h, function DE0C requires that esi  must contain a linear address
  ; pointing here.
                          ;NOTE: this Data MUST be in LOW_MEM (below 1 Mbyte),
                          ;      data referenced by this structure
                          ;      CAN be in memory above 1 Mbyte
                          ;      On switching to protected mode the server
                          ;      first loads CR3 (paging base) from "our_CR3".
                          ;

                        ; value of CR3 to be loaded by server

; The ones initialized to zero get computed in _exit_vm86

    our_CR3    dd       0
                        ; linear address in first Mbyte pointing to
                        ; value of GDTR ("our_GDTR") to be loaded by server
    our_GDTRptr dd      0
                        ; linear address in first Mbyte pointing to
                        ; value of IDTR ("our_IDTR") to be loaded by server
    our_IDTRptr dd      0
                        ; value of LDTR to be loaded by server
    our_LDTR   dw       g_ldt
                        ; value of TR to be loaded by server
    our_TR     dw       g_tss
                        ; Fword, pointer to code to be started by server
    protected_mode_target DD ?
                          DW g_code
  ;  belongs to above
    our_GDTR   dw       (gdtlast-gdtnull)-1    ;limit (byte gran)
                        ;linear (not physical) base address of "gdtnull"
    laddr_GDT  dd       0
               dw       ?    ; (just for align "laddr_IDT" to Dword)
    our_IDTR   dw       (idtlast -idtnull) - 1
                        ;linear (not physical) base address of "Idtnull"
    laddr_IDT  dd       0
  ;-------------------------------------------------------------------^

  server_vcpi_entry df  0 ; this is the address we must call instead of INT67
                          ; when in protected mode
  pagedir_template dd    0 

  local_storage dd ?


pmode_return_esp   dd   ?  ; used to return from real-mode-switch
real_IDTR          dw       03ffh    ;needed to load real intvector after switch
                   dd       0        ; lidt loads 6 bytes into the idtr.
                                     ; The last 4 (=0) gives base address
                                     ; of the real mode vector table


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                           ;
;                             END OF DATA                                   ;
;                  (Well, there's a bit at the end too.)                    ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; The following routine returns 1 or 0 according as VCPI services
; are (are not) present.
;
; Calling sequence from C: int check_VCPI_present();
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ASSUME ds:DGROUP       ; We need to access data in the c-module

_check_VCPI_present proc far
          pushad
          push    es
          push    ds
          xor     eax,eax      ; check EMM_interrupt present
          mov     return_value,ax ; default return value is 0 (= NO)
          mov     es,ax
          mov     es,word ptr es:[emm_int*4+2]
          cmp     Dword ptr es:[10+4],'0XXX'
          jne     @@ex
          cmp     dword ptr es:[10],'XMME'
          je      @@ok
          cmp     dword ptr es:[10],'QMME' ; this also works (as told by <J.S.Peatfield@damtp.cambridge.ac.uk>)
          jne     @@ex
@@ok:
          mov     ax,4000h   ; check emm manager status
          int     emm_int
          or      ah,ah
          jnz     @@ex
          mov     ax,_GET_VERSION*256   ; get version
          int     emm_int
          cmp     ax,40h     ; version must be greater 4.0
          jb      @@ex

          mov     ax,_ALLOCATE_RAW_PAGES
          mov     bx,4
          int     emm_int
          cmp     ah,0
          jnz     @@ex
          mov     handle,dx     ; save handle
          mov     ax,0DE00h   ; check for VCPI present
          int     emm_int
          cmp     ah,0
          jne     short @@notok
          mov     return_value,1 ; return value is 1 (=YES)
@@notok:
          mov     dx, handle
          mov     ax,_DEALLOCATE_PAGES*256
          int     emm_int
@@ex:
          pop     ds
          pop     es
          popad
          mov      ax, return_value
          ret
_check_VCPI_present endp

ASSUME ds:CODE   ; From now on, all data is accessed from cs

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                          ;
; get_VCPI_interface: front-end for get_protected_mode_interface           ;
;                                                                          ;
; Preserves all registers, returns nothing.                                ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

get_VCPI_interface proc near
          push    cs                  ;* buffer for server GDT entries
          push    offset gdtvcpi_code ;*
          push    cs                  ;* buffer for server page0
          push    page0        ;
          call    get_protected_mode_interface
          ret
get_VCPI_interface endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                          ;
; get_protected_mode_interface: Front-end for int 67h service DE01h        ;
;                                                                          ;
; This is used to prepare for a switch to protected mode using the VCPI    ;
; server. The above-named service fills three descriptors in the GDT used  ;
; by the server when running in protected mode. It also fills a buffer with;
; the server's current page table. (Cf sections 11.3 and 11.4 of Vol 3. of ;
; Pentium Developer's manual for information about paging in protected     ;
; mode. We assume throughout this code that paging is enabled in order for ;
; the server to provide extended memory blocks, but that the mapping of    ;
; conventional memory is such that linear addresses = physical addresses.) ;
;
; Before calling, the following must be pushed on the stack:
;  1) A word giving the offset in the current code segment of the start of
;      3 contiguous unitialized descriptors in what will become the GDT after
;      the switch to protected mode.
;
;  2) A word giving the offset in the current code segment of the beginning
;     of a buffer for the server's page table. The address pointed to must
;     be a multiple of 4096, and the buffer must hold 4096 bytes.
;
;  This routine uses the Pascal calling convention.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

get_protected_mode_interface proc near
          ARG     server_pagebuffer:dword,server_selectors:dword = returnpop
          push    bp
          mov     bp,sp
          push    si
          push    di
          push    ds
          push    es

                                    ; convert pagebuffer to linear address
          movzx   eax,word ptr server_pagebuffer+2
          movzx   ebx,word ptr server_pagebuffer
          shl     eax,4
          add     eax,ebx
          shr     eax,4      ; make real mode address

          ;XXXXXXXXXX
          push    ds
          mov     es,ax                ; ES = frame of server_pagebuf
          xor     di,di                ; ES:DI = >buffer for page0-table<
          lds     si,server_selectors ; DS:SI = >3 GDT entries for VCPI
                               ; calling VCPI server
          mov     ax,0DE01h
          int     emm_int
          cmp ah,0          ; Test to see if it worked
	  jz AOK
	  debugout mystring6
AOK:
          pop     ds

                                   ; save VCPI server entry for protected mode
          mov     dword ptr server_vcpi_entry,ebx
                                      ; set the VCPI-Code selector
          mov     word ptr server_vcpi_entry+4,g_vcpi_code
          ;XXXXXXXXXXX



          pop     es
          pop     ds
          pop     di
          pop     si
          pop     bp
          ret     returnpop ; PASCAL calling convention. Cleans up stack
                            ; on return to caller
get_protected_mode_interface endp


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                           ;
; pmode_task: run as a task by the VCPI server after switch to protected    ;
;  mode using int 67h. (Written by Hans Lerman)                            ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pmode_task proc near
; We come here in protected mode via VCPI-CALL
; have only to set up all things to switch back to 386-realmode
;
; We follow the INTEL recommended procedure to do this
; (Pentium Processor Family Developer's Manual, Vol 3., Section 16.6.2)
;
          .386p
          CLI  ; just to be sure. No need to set up new stack.

          ; 1. We know: that paging is enabled,
          ;             that linear addresses have identical mapping
          ;    so we can clear the PG bit

          mov     eax,CR0
          and     eax,07fffffffh
          mov     CR0,eax    ; paging is off
          xor     eax,eax
          mov     CR3,eax    ; clear out paging cache

          ; 2. We must transfer control to a segment with 64K limit
          ;    we ARE on such a segment, no need to do it here

          ; 3. Load segment registers SS,DS,ES,FS,GS with a selector
          ;    that points to a descriptor containing the values:
          ;      limit 64k     =  0FFFFh
          ;      byte granularity
          ;      Expand up
          ;      writable
          ;      present
          ;
          ;    We have that descriptor, it's our gdt_data.
          mov     ax,g_data
          mov     ss,ax
          mov     ds,ax
          mov     es,ax
          mov     fs,ax
          mov     gs,ax


          ; 4. Disable Interrrupts
          ;    ( we did that above )
          ; 5. Clear the PE bit in CR0
          mov     eax,CR0
          and     al,0feh
          mov     CR0,eax  ; we are going back to realmode

          ; 6. Flush the instruction queue and load CS with appropriate value
          ;jmp     far ptr @@here
          db      0EAh ; machine code for far jump
          dw      @@here
reloc_hseg_1  dw  0    ; 0 gets replaced from _exit_vm86
@@here    label   far

          ; 7. Use LIDT instruction to load the base and limits of the
          ;    real-mode interrupt vector table
          lidt    fword ptr cs:real_IDTR

          .386
          ; 8. Enable interrupts (we are doing this later in _exit_vm86)
          ; 9. Load segment register
          mov     ax,cs:C_SEG
          mov     ds,ax
          mov     es,ax
          mov     fs,ax
          mov     gs,ax
          mov     esp,pmode_return_esp
	  mov     ss, S_SEG
          ; now we are in "real" real-mode
          popad
          ret
pmode_task endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                          ;
; switch_to_protected_mode_and_return_in_386realmode: well, the name says  ;
; it all, doesn't it?                                                      ;
;                                                                          ;
; NOTE: must have called get_VCPI_interface before this                    ;
;                                                                          ;
; returns with interrupts disabled                                         ;
; all registers preserved, all segment registers = "C_SEG"                 ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

switch_to_protected_mode_and_return_in_386realmode proc near
          
          pushad
          mov    pmode_return_esp,esp
          lea    ax,pmode_task
          movzx  eax,ax
          mov    dword ptr protected_mode_target,eax
          mov    word ptr protected_mode_target+4,g_code
	  mov ax, offset our_CR3
	  movzx eax, ax
          add eax, C_FRAME
          mov     esi,eax  ; linear address

	debugout mystring4

          CLI
          mov     ax,0DE0Ch
          int     emm_int    ; jumps to protected mode
          ; does NOT return here
          ; "pmode_task" does return directly to our caller
	  debugout mystring5  ; so this should never happen
switch_to_protected_mode_and_return_in_386realmode endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                                                                           ;
; _exit_vm86: The main routine of this module. Does exactly what the name   ;
;   says.                                                                   ;
;                                                                           ;
; Call from C as: exit_vm86();   /* And hope for the best :-) */            ;    ;
;                                                                           ;
; Note: after calling this, expect all DOS services to be HOSED.            ;
;                                                                           ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

_exit_vm86 proc far

	; copy value of debug flag from C-module for local use by
        ; debugout macro
	mov ax, word ptr _debug
	mov cs:debug_flag, ax

        ; load ds register: it currently points at c-module data.
	mov cs:D_SEG,ds
	mov ax,cs
	mov ds,ax

        ; Many things need to get initialized below using the current value
        ; of cs and cs << 4 (i.e, frame). We store these for later use now.
        mov C_SEG, ax
        movzx eax,ax   
	shl eax,4
        mov C_FRAME, eax  ; Used for converting logical addresses to linear

	debugout mystring0

        ; finish initialization of GDT by filling in base address fields
        fixup_descriptor gdt_code,C_FRAME    ; our code segment descriptor
        fixup_descriptor gdt_data,C_FRAME    ; our data segment descriptor

        ; fixup our ldt descriptor
	mov ax, (ldtnull-code_org0)
	movzx eax,ax
        add eax, C_FRAME
	mov local_storage, eax
	fixup_descriptor gdt_ldt,local_storage

        ; fixup our task state segment descriptor
        mov ax,(our_tss - code_org0)
	movzx eax,ax
        add eax, C_FRAME
	mov local_storage, eax
	fixup_descriptor gdt_tss,local_storage

        ; Finish initialization of protected mode params
        ;
        ; These are used by the int 67h DE0Ch service. In particular, the
        ; server wants the location of the page directory, whose linear
        ; address it will load in the CR3 register. We store this for
        ; the server in our_CR3. We store a (near) pointer to the page directory
        ; in a variable called pagedir, so we must load the contents of
        ; pagedir into our_CR3 after converting to linear form. The actual
        ; beginning of the page table is computed and pointed to by a near
        ; pointer stored in page0.

        ; The reason we must compute the addresses of the page table and
        ; of the page directory (which is, itself, a page table) at run time
        ; is that
        ; page tables must be aligned at multiples of 4096. Since we have
        ; no idea where DOS will load our program, we cannot compute the
        ; addresses of these tables at assembly time.

	; compute location for pagedir
        ; and for page0.
	; These must be aligned at 4096.
	; They will land somewhere inside buffer1
	push ebx
	push edx
	mov ebx, 4096
	xor edx,edx
	mov ax, offset buffer1
	movzx eax,ax
        add eax, C_FRAME ; eax now has linear address of buffer1
        div ebx          ; (x/4096 +1)*4096 rounds x to next multiple of 4096
	inc eax
	mul ebx
        sub eax,C_FRAME   ; back to logical address
        mov pagedir,ax    ; and store in pagedir
	add eax,ebx
        mov page0,ax      ; store address of next mempage in page0
	pop edx
	pop ebx

        mov ax, pagedir   ; convert address in pagedir to linear and store
        movzx eax,ax      ; in our_CR3 for server to load in CR3
        add eax, C_FRAME
	mov our_CR3, eax

        ; Although the server will fill in the page table, it is our
        ; responsibility to fill in the page directory (whose address was
        ; computed above). In this case, the page directory has only one
        ; entry: a linear address pointing to the page table whose address-3
        ; is stored in page0

	mov ax, page0 
	movzx eax,ax
	add eax,3
        add eax, C_FRAME
	mov pagedir_template, eax

	push di
	push si
	mov ax,ds
	mov es,ax
	mov ax, offset pagedir_template
	mov si,ax
	mov ax,pagedir  ; note: contents, not offset
	mov di,ax
        movsw        ; copy 4 words starting at pagedir_template 
        movsw        ; to place pointed by pagedir
	pop si
	pop di

        ; The server wants a linear pointer to an address containing a value
        ; it will use to load into the gdtr ( This address, our_GDTR, was
        ; loaded at assembly time
	mov ax,  (our_GDTR - code_org0)
	movzx eax,ax
        add eax,C_FRAME
	mov our_GDTRptr, eax

        ; Same thing for the idtr
	mov ax,  (our_IDTR-code_org0)
	movzx eax,ax
        add eax, C_FRAME
	mov our_IDTRptr, eax

        ; Next, the server wants to know the location of the GDT we will use
        ; in protected mode. We store the linear address in laddr_GDT

	mov ax, (gdtnull - code_org0)
	movzx eax,ax
        add eax, C_FRAME
	mov laddr_GDT,eax

        ; And same for the IDT:
	mov ax, (idtnull-code_org0)
	movzx eax,ax
        add eax, C_FRAME
	mov laddr_IDT ,eax

        ; There is a peculiar long jump command in the task run from
        ; protected mode. The only purpose of this command is to force
        ; a load of cs for the return to real mode. We had to leave the
        ; segment address of that jump blank at assembly time, so we
        ; fix that up now.

        mov ax, C_SEG
        mov reloc_hseg_1, ax ; store cs above for far jump
        
        mov E_SEG, es  ; store so they can be restored after return
	mov S_SEG, ss


	debugout mystring1

        call get_VCPI_interface

	debugout mystring2


        call switch_to_protected_mode_and_return_in_386realmode

done:
        mov es, E_SEG       ; Stack seg already restored above
        mov ds, D_SEG
        ret
_exit_vm86 endp


pagedir dw  (?)     ; pointer to 4096-aligned mempage inside buffer1
page0   dw  (?)     ; pointer to following mempage inside buffer1
buffer1 dd 3072 dup(?)  ; big enough to ensure we can fit the two mempages

CODE ENDS
END
